# Generated by Django 4.1.7 on 2023-05-07 15:38

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("gqlauth", "0001_initial"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="userstatus",
            options={"verbose_name_plural": "User statuses"},
        ),
    ]
