# Generated by Django 4.1.9 on 2023-05-07 15:59

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("gqlauth", "0002_alter_userstatus_options"),
    ]

    operations = [
        migrations.DeleteModel(
            name="Captcha",
        ),
    ]
